// Copyright (c) 2012, Alexandre Ardhuin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

library js_wrap;

import 'dart:collection';

import 'package:meta/meta.dart';
import 'package:js/js.dart' as js;

import 'optional.dart';

class ProxyInvocationMirror extends InvocationMirror {
  final String memberName;
  List positionalArguments;
  final Map<String, dynamic> namedArguments;
  final bool isMethod;
  final bool isGetter;
  final bool isSetter;
  ProxyInvocationMirror(this.memberName, List positionalArguments, this.namedArguments, this.isMethod, this.isGetter, this.isSetter) {
    this.positionalArguments = positionalArguments != null ? positionalArguments.mappedBy(_serialize).toList() : null;
  }
  ProxyInvocationMirror.fromInvocationMirror(InvocationMirror invocation) : this(invocation.memberName, invocation.positionalArguments, invocation.namedArguments, invocation.isMethod, invocation.isGetter, invocation.isSetter);
  ProxyInvocationMirror.method(String memberName, List positionalArguments) : this(memberName, positionalArguments, {}, true, false, false);
  ProxyInvocationMirror.getter(String memberName) : this(memberName, [], {}, false, true, false);
  ProxyInvocationMirror.setter(String memberName, dynamic value) : this(memberName, [value], {}, false, false, true);
  @override invokeOn(Object receiver) { throw new UnsupportedError("Forbidden"); }
}

typedef Object Transformater(Object proxy);

// transform IsJsProxy to js.Proxy
Object _serialize(data) =>
    (data is IsJsProxy) ? data._jsProxy._proxy :
    (data is IsEnum) ? data.value :
    (data is List) ? js.array(data.mappedBy(_serialize).toList()) :
    (data is Date) ? new js.Proxy(js.context.Date, data.millisecondsSinceEpoch) :
    data;

class JsProxy {
  js.Proxy _proxy;

  JsProxy._(this._proxy);

  @override noSuchMethod(InvocationMirror invocation) {
    final proxyInvocation = new ProxyInvocationMirror.fromInvocationMirror(invocation);
    final jsResult = _proxy.noSuchMethod(proxyInvocation);
    // print("${method}(${args}) => ${jsResult}");
    return asOption(jsResult);
  }
}

/// Base class of object that wraps jsProxy
class IsJsProxy {
  JsProxy _jsProxy;

  IsJsProxy() : this.fromJsProxy(js.map({}));
  IsJsProxy._(this._jsProxy);
  IsJsProxy.fromIsJsProxy(IsJsProxy isJsProxy) : this._(isJsProxy._jsProxy);
  IsJsProxy.fromJsProxy(js.Proxy proxy) : this._(new JsProxy._(proxy));
  IsJsProxy.newInstance(objectRef, [List args]) : this.fromJsProxy(new js.Proxy.withArgList(objectRef, args != null ? args.mappedBy(_serialize).toList() : []));

  JsProxy get $ => _jsProxy;
}

/// retain [isJsProxy]
IsJsProxy retain(IsJsProxy isJsProxy) {
  js.retain(isJsProxy._jsProxy._proxy);
  return isJsProxy;
}
/// retain all elements in [isJsProxyList]
void retainAll(List<IsJsProxy> isJsProxyList) {
  isJsProxyList.forEach(retain);
}
/// release [isJsProxy]
void release(IsJsProxy isJsProxy) {
  js.release(isJsProxy._jsProxy._proxy);
}

class JsDate extends IsJsProxy implements Date {
  static final INSTANCIATOR = (js.Proxy jsProxy) => new JsDate.fromJsProxy(jsProxy);

  JsDate.fromJsProxy(js.Proxy proxy) : super.fromJsProxy(proxy);

  // from Date->Comparable
  @override int compareTo(Date other) => _asDate().compareTo(other);

  // from Date
  @override bool operator ==(Date other) => _asDate() == other;
  @override bool operator <(Date other) => _asDate() < other;
  @override bool operator <=(Date other) => _asDate() <= other;
  @override bool operator >(Date other) => _asDate() > other;
  @override bool operator >=(Date other) => _asDate() >= other;
  @override Date toLocal() => _asDate().toLocal();
  @override Date toUtc() => _asDate().toUtc();
  @override String get timeZoneName => _asDate().timeZoneName;
  @override Duration get timeZoneOffset => _asDate().timeZoneOffset;
  @override int get year => _asDate().year;
  @override int get month => _asDate().month;
  @override int get day => _asDate().day;
  @override int get hour => _asDate().hour;
  @override int get minute => _asDate().minute;
  @override int get second => _asDate().second;
  @override int get millisecond => _asDate().millisecond;
  @override int get weekday => _asDate().weekday;
  @override int get millisecondsSinceEpoch => _asDate().millisecondsSinceEpoch;
  @override bool get isUtc => _asDate().isUtc;
  @override String toString() => _asDate().toString();
  @override Date add(Duration duration) => _asDate().add(duration);
  @override Date subtract(Duration duration) => _asDate().subtract(duration);
  @override Duration difference(Date other) => _asDate().difference(other);

  Date _asDate() => new Date.fromMillisecondsSinceEpoch($.getTime().value);
}

class _JsIterator<E> implements Iterator<E> {
  final JsList<E> _jsList;
  final int length;
  int _currentIndex = -1;

  _JsIterator(JsList<E> jsList) : this._jsList = jsList, length = jsList.length;

  // Iterator
  @override bool moveNext() {
    if (_currentIndex + 1 < length) {
      _currentIndex++;
      return true;
    }
    return false;
  }
  @override E get current => _jsList[_currentIndex];
}

class JsList<E> extends IsJsProxy implements List<E> {
  Transformater _instantiator;

  JsList.fromJsProxy(js.Proxy proxy, [E instantiator(e)]) : super.fromJsProxy(proxy) {
    _instantiator = instantiator != null ? instantiator : (e) => e;
  }

  // Object
  @override String toString() => _asList().toString();

  // Iterable
  @override Iterator<E> get iterator => new _JsIterator<E>(this);
  @override int get length => $.length.value;

  // Collection
  @override void add(E value) { $.push(value); }
  @override void clear() { $.splice(0, length); }
  @override void remove(Object element) { removeAt(indexOf(element)); }

  // List
  @override E operator [](int index) => $[index].map(_instantiator).value;
  @override void operator []=(int index, E value) { $[index] = value; }
  @override void set length(int newLength) {
    final length = this.length;
    if (length < newLength) {
      final nulls = new List<E>(newLength - length);
      addAll(nulls);
    }
    if (length > newLength) {
      throw new UnsupportedError("New length has to be greater than actual length");
    }
  }
  @override void addLast(E value) { $.push(value); }
  @override List<E> get reversed => _asList().reversed;
  @override void sort([int compare(E a, E b)]) {
    final sortedList = _asList()..sort(compare);
    clear();
    addAll(sortedList);
  }
  @override int indexOf(E element, [int start = 0]) => _asList().indexOf(element, start);
  @override int lastIndexOf(E element, [int start]) => _asList().lastIndexOf(element, start);
  @override E removeAt(int index) => ($.splice(index, 1).map((proxy) => new JsList<E>.fromJsProxy(proxy, _instantiator)).value as JsList<E>)[0];
  @override E removeLast() => $.pop().map(_instantiator).value;
  @override List<E> getRange(int start, int length) => _asList().getRange(start, length);
  @override void setRange(int start, int length, List<E> from, [int startFrom = 0]) {
    final args = [start, 0];
    for(int i = startFrom; i < length; i++) {
      args.add(from[i]);
    }
    $.noSuchMethod(new ProxyInvocationMirror.method("splice", args));
  }
  @override void removeRange(int start, int length) { $.splice(start, length); }
  @override void insertRange(int start, int length, [E initialValue]) {
    final args = [start, 0];
    for (int i = 0; i < length; i++) {
      args.add(initialValue);
    }
    $.noSuchMethod(new ProxyInvocationMirror.method("splice", args));
  }

  List<E> _asList() {
    final list = new List<E>();
    for (int i = 0; i < length; i++) {
      list.add($[i].value);
    }
    return list;
  }

  // -------remove this copy of Iterable when mixin will land
  @override Iterable map(f(E element)) => IterableMixinWorkaround.map(this, f);
  @override Iterable mappedBy(f(E element)) => IterableMixinWorkaround.map(this, f);
  @override Iterable<E> where(bool f(E element)) => IterableMixinWorkaround.where(this, f);
  @override bool contains(E element) => IterableMixinWorkaround.contains(this, element);
  @override void forEach(void f(E element)) => IterableMixinWorkaround.forEach(this, f);
  @override dynamic reduce(var initialValue, dynamic combine(var previousValue, E element)) => IterableMixinWorkaround.reduce(this, initialValue, combine);
  @override bool every(bool f(E element)) => IterableMixinWorkaround.every(this, f);
  @override String join([String separator]) => IterableMixinWorkaround.join(this, separator);
  @override bool any(bool f(E element)) => IterableMixinWorkaround.any(this, f);
  @override List<E> toList() => new List<E>.from(this);
  @override Set<E> toSet() => new Set<E>.from(this);
  @override E min([int compare(E a, E b)]) => IterableMixinWorkaround.min(this, compare);
  @override E max([int compare(E a, E b)]) => IterableMixinWorkaround.max(this, compare);
  @override bool get isEmpty => IterableMixinWorkaround.isEmpty(this);
  @override Iterable<E> take(int n) => IterableMixinWorkaround.takeList(this, n);
  @override Iterable<E> takeWhile(bool test(E value)) =>IterableMixinWorkaround.takeWhile(this, test);
  @override Iterable<E> skip(int n) => IterableMixinWorkaround.skipList(this, n);
  @override Iterable<E> skipWhile(bool test(E value)) => IterableMixinWorkaround.skipWhile(this, test);
  @override E get first => IterableMixinWorkaround.first(this);
  @override E get last => IterableMixinWorkaround.last(this);
  @override E get single => IterableMixinWorkaround.single(this);
  @override E firstMatching(bool test(E value), { E orElse() }) => IterableMixinWorkaround.firstMatching(this, test, orElse);
  @override E lastMatching(bool test(E value), {E orElse()}) => IterableMixinWorkaround.lastMatching(this, test, orElse);
  @override E singleMatching(bool test(E value)) => IterableMixinWorkaround.singleMatching(this, test);
  @override E elementAt(int index) => IterableMixinWorkaround.elementAt(this, index);

  // Collection
  @override void addAll(Iterable<E> elements) {
    for (E element in elements) {
      add(element);
    }
  }
  @override void removeAll(Iterable elements) => IterableMixinWorkaround.removeAll(this, elements);
  @override void retainAll(Iterable elements) => IterableMixinWorkaround.retainAll(this, elements);
  @override void removeMatching(bool test(E element)) => IterableMixinWorkaround.removeMatching(this, test);
  @override void retainMatching(bool test(E element)) => IterableMixinWorkaround.retainMatching(this, test);
}

class IsEnum<E> {
  E value;

  IsEnum(this.value);
}

Object findIn(Object o, List<Object> elements, [Object transform(Object)]) {
  final matchingElements = elements.where((e) => ((transform != null ? transform(e) : e) == o));
  if (matchingElements.length == 1) {
    return matchingElements.first;
  } else {
    return null;
  }
}

class Callback extends js.Callback {
  Callback.once(Function f) : super.once(_serializeResult(f));
  Callback.many(Function f) : super.many(_serializeResult(f));
  static _serializeResult(f) => ([arg1, arg2, arg3, arg4]) {
    // TODO check number of parameter to avoid:"Closure argument mismatch"
    // for exemple Events.addDomListener(mapDiv, 'click', () { window.alert('DIV clicked'); });
    // callback has no interest for event param
    if (!?arg1) {
      return _serialize(f());
    } else if (!?arg2) {
      return _serialize(f(asOption(arg1)));
    } else if (!?arg3) {
      return _serialize(f(asOption(arg1), asOption(arg2)));
    } else if (!?arg4) {
      return _serialize(f(asOption(arg1), asOption(arg2), asOption(arg3)));
    } else {
      return _serialize(f(asOption(arg1), asOption(arg2), asOption(arg3), asOption(arg4)));
    }
  };
}